/*{{{  Includes*/
#include "juggle.h"
#include "x_uif.h"
#include "3d.h"
#include "animate.h"
/*}}}  */
/*{{{  Typedefs*/
typedef struct { int   x,y; } point2d;
/*}}}  */
/*{{{  Globals*/
#define ITEM_WIDTH  12
#define ITEM_HEIGHT 12

#define   MAX_COLOR_BLTS 16

int       MyItemIndex   = 0;
int       MyMaxColor;
Pixmap    pixmaps[MAX_COLOR_BLTS];
Pixel     colors[MAX_COLOR_BLTS];
Pixel     fgPixel, bgPixel;

char      drawn[MAXNV][MAXNV];
/*}}}  */

void
map_colours(Pixel *fg, Pixel *bg)
/*{{{  */
{
   Display  *disp    = xdata.display;
   int      screen   = xdata.screen;
   Window   win      = DefaultRootWindow(disp);
   Colormap colormap = DefaultColormap(disp, screen);
   int      depth    = DefaultDepth(disp, screen);
   int      visid    = DefaultVisual(disp,screen)->class;
   int      max, col, stat_col;
   
   XSetWindowColormap(disp, win, colormap);
   
   xdata.depth = depth;
   xdata.mono  = (depth == 1);

   /*{{{  Basic colours*/
   *fg = fgPixel = BlackPixel(disp, screen); 
   *bg = bgPixel = WhitePixel(disp, screen);
   /*}}}  */
      
   /*{{{  GCs*/
   xdata.XorGC = XCreateGC(disp, win, None, NULL);
   XSetBackground(disp, xdata.XorGC, bgPixel);
   XSetForeground(disp, xdata.XorGC, fgPixel);
   XSetGraphicsExposures(disp, xdata.XorGC, False);
   XSetFunction(disp, xdata.XorGC, ( (bgPixel&1) == 1 ) ? GXequiv : GXxor);
   
   xdata.RevGC = XCreateGC(disp, win, None, NULL);
   XSetBackground(disp,xdata.RevGC, fgPixel);
   XSetForeground(disp,xdata.RevGC, bgPixel);
   XSetGraphicsExposures(disp, xdata.RevGC, False);
   
   xdata.ColorGC = XCreateGC(disp, win, None, NULL);
   XSetBackground(disp,xdata.ColorGC, bgPixel);
   XSetForeground(disp,xdata.ColorGC, fgPixel);
   XSetGraphicsExposures(disp, xdata.ColorGC, False);
   /*}}}  */
   
   /*{{{  Allocate random colors, or use pre-existing ones*/
   MyMaxColor = 0;
   max = min(MAX_COLOR_BLTS, 1<<depth); 
   stat_col = ((visid == StaticGray) || (visid == StaticColor));
   
   for(col=0; col<max; col++)
   {
      XColor   color;
      long     cells[10];
   
      color.flags = DoRed | DoGreen | DoBlue;
      
      if (!stat_col)
      {
         if (!XAllocColorCells (disp, colormap,
	                        TRUE,    /* contiguous */
	                        NULL, 0, /* no planes */
                                cells, 1)/* 1 cell at a time */ ) 
         {
            break; /* Can't allocate any more colors */
         }
   
         color.red   = rand()%65535;
         color.green = rand()%65535;
         color.blue  = rand()%65535;
         color.pixel = cells[0];
         XStoreColor(disp, colormap, &color);
         
         colors[col] = color.pixel;
      }
      else
         colors[col] = rand()%(1<<depth);
   
      if (colors[col] != bgPixel)
      {
         /*{{{  Create static pixmaps for balls*/
            /* Create pixmap for item */
         pixmaps[col] = XCreatePixmap(disp, win,
                                      ITEM_WIDTH, ITEM_HEIGHT,
                                      depth);
         
         /* Clear pixmap */
         XFillRectangle(disp,
                        pixmaps[col],
                        xdata.RevGC,
                        0, 0,
                        ITEM_WIDTH, ITEM_HEIGHT);
         
         /* Draw item */
         XSetForeground(disp, xdata.ColorGC, colors[col]);
         
         XFillArc(disp,
                  pixmaps[col],
                  xdata.ColorGC,
                  0, 0,
                  ITEM_WIDTH-1,
                  ITEM_HEIGHT-1,
                  0, 360*64);
            
         XSetForeground(disp, xdata.ColorGC, bgPixel);
         
         XFillArc(disp,
                  pixmaps[col],
                  xdata.ColorGC,
                  ITEM_WIDTH/4, ITEM_HEIGHT/8,
                  3*ITEM_WIDTH/8, 3*ITEM_HEIGHT/8,
                  0, 360*64);
         /*}}}  */
         
         MyMaxColor++;
      }
   }
   /*}}}  */
} 
/*}}}  */

int
map2d(point *p, int *x, int *y)
/*{{{  */
{
   /* Perform a 3d -> 2d mapping */
   
   
   if (p->z >= ds)
   {
      printf("Invisible point - (%d,%d,%d)\n",p->x,p->y,p->z);
   
      *x = p->x;
      *y = p->y;
   }
   else
   {
      *x = (abs(de) * p->x) / (de + ds - p->z);
      *y = (abs(de) * p->y) / (de + ds - p->z);
   }

  /* Move the origin to the centre of the virtual window */
  
   *x += STD_WIDTH2;
   *y += STD_HEIGHT2;
  
  /* 
   * And scale the virtual coords to the actual window size.
   *
   * What this does is actually x' = round(x/STD_WIDTH * cur_width)
   *  (and similar for y)
   */

   *x = ((*x-STD_WIDTH2) * cur_width) / STD_WIDTH + cur_width/2;
  
   *y = STD_HEIGHT-*y;
   *y = ((*y-STD_HEIGHT2) * cur_height) / STD_HEIGHT + cur_height/2;

   return 1;
}
/*}}}  */

void
do_draw_line(point *p1, point *p2, GC gc)
/*{{{  */
{
   int   x1,y1,x2,y2;
   point *b, *f, clipp;
   int   clipz = min(ds, ds+de);
   
   if (p1->z > p2->z)
   {
      b = p1; f = p2;
   }
   else
   {
      b = p2; f = p1;
   }
   
   if (f->z >= clipz)
      return;       /* Completely invisible */
      
   if (b->z < clipz)
      ;   /* Completely visible  */
   else
   {                /* Partially visible   */
      clipz -= clip_delta;
      
      clipp.x = b->x + ((b->z - clipz) * f->x)/(b->z - f->z);
      clipp.y = b->y + ((b->z - clipz) * f->y)/(b->z - f->z);
      clipp.z = clipz;
      
      /*
      printf("Clipping {%d,%d,%d} {%d,%d,%d} clipz=%d -> {%d,%d,%d}\n",
                  b->x,b->y,b->z,
                  f->x,f->y,f->z,
                  clipz,
                  clipp.x,clipp.y,clipp.z);
      */
      b = &clipp;
   }
   
   if ( map2d(b,&x1,&y1) && map2d(f,&x2,&y2) )
   {
      /*
      if (b == &clipp)
         printf("   Mapping -> (%d,%d)(%d,%d)\n",x1,y1,x2,y2);
      */
      XDrawLine(xdata.display,
                xdata.drawWin,
                gc,
                x1,y1,
                x2,y2);
   }
}
/*}}}  */

void
draw_line(point *p1, point *p2)
/*{{{  */
{
   point vp1,vp2;
   
   tx_point(p1, &view_tx, &vp1);
   tx_point(p2, &view_tx, &vp2);
   
   XSetForeground(xdata.display, xdata.XorGC, fgPixel);
   do_draw_line(&vp1, &vp2, xdata.XorGC);
}
/*}}}  */

void
update_ap(anim_pix *ap, Polyinfo *poly, int ball_i)
/*{{{  */
{
   pix *bp  = ap->pix;
   tx  *btx = ap->tx;
   
   if (poly == NULL)
      poly = prop_index < 0 ? &props[ball_i % props_size] :
                              &props[prop_index];
      
   if (clubs || ball_i < 0)
   {
      tx  jtx;
      int      *face = &poly->f[0];
      point    pts[MAXVERTS];
      point2d  pts2[MAXVERTS];
      point2d  ul = {STD_WIDTH+1,STD_HEIGHT+1},
               lr = {-1,-1};
      int      w,h;
      XPoint   xpts[MAXEDGESPERPOLY];
      int      i,j,p0,p1,x,y;
      int      clipz = min(ds, ds+de);
      
      multiply_tx(btx, &view_tx, &jtx);
      
      /*{{{  Transform & map all vertices, find ll, ur points*/
      for(i=0; i < poly->numverts; i++)
      {
         tx_point(&poly->v[i], &jtx, &pts[i]);
         
         if (pts[i].z >= clipz)   /* Some point of obj is behind eye */
         {
            bp->w = bp->h = 0;
            return;
         }
            
         map2d(&pts[i], &pts2[i].x, &pts2[i].y);
         ul.x = min(ul.x, pts2[i].x);
         ul.y = min(ul.y, pts2[i].y);
         lr.x = max(lr.x, pts2[i].x);
         lr.y = max(lr.y, pts2[i].y);
      }
      /*}}}  */
      
      w = lr.x - ul.x;
      h = lr.y - ul.y;
      
      if (w==0 || h==0) return;
      
      /*{{{  Update pixmap*/
      fit_pix(bp, w, h);
      
      bp->w = w;
      bp->h = h;
      bp->x = ul.x;
      bp->y = ul.y;
      /*}}}  */
      
      for(i=0; i < poly->numverts; i++)
         memset(&drawn[i], 0, poly->numverts);
         
      XFillRectangle(xdata.display, bp->pm, xdata.RevGC, 0, 0, w, h);
      
      /*{{{  Draw the polyhedron into the pixmap*/
      for(i=0; i < poly->numfaces; i++)
      {
         int verts  = face[0];
         int area   = 0;
         
         /*{{{  Calculate clockwiseness*/
         for(j=0; j<verts; j++)
         {
            p0 = face[1 + j];
            p1 = face[1 + (j+1)%verts];
            area += (pts2[p1].x - pts2[p0].x)*(pts2[p1].y + pts2[p0].y);
         }
         /*}}}  */
         
         if (area < 0)
         {
            int col = xdata.mono ? fgPixel :
                      ball_i < 0 ? colors[(1-ball_i)%MyMaxColor] :
                                   colors[(ball_i+i)%MyMaxColor];
            
            XSetForeground(xdata.display, xdata.XorGC, col);
            
            if (fill)
            {
               for(j=0; j<verts; j++)
               {
                  p0 = face[1+j];
                  xpts[j].x = pts2[p0].x - ul.x;
                  xpts[j].y = pts2[p0].y - ul.y;
               }
               
               XFillPolygon(xdata.display, 
                            bp->pm,
                            xdata.XorGC,
                            xpts, verts,
                            Complex, CoordModeOrigin);
                                                            
               XSetForeground(xdata.display, xdata.XorGC, fgPixel);
            }
            
            for(j=0; j<verts; j++)
            {
               p0 = face[1 + j];
               p1 = face[1 + (j+1)%verts]; 
               
               if (!drawn[p0][p1])
               {
                  drawn[p0][p1] = drawn[p1][p0] = 1;
                  XDrawLine(xdata.display,
                            bp->pm,
                            xdata.XorGC,
                            pts2[p0].x - ul.x, pts2[p0].y - ul.y,
                            pts2[p1].x - ul.x, pts2[p1].y - ul.y);
               }
            }
         }
         
         face += verts+1;
      }
      /*}}}  */
   }
   else
   {
      point p = {0,0,0};
      point vp;
      int   x,y;
       
      tx_point(&p, btx, &p);
      tx_point(&p, &view_tx, &vp);
      map2d(&vp, &x, &y);
      
      fit_pix(bp, ITEM_WIDTH, ITEM_HEIGHT);
      bp->w = ITEM_WIDTH;
      bp->h = ITEM_HEIGHT;
      bp->x = x - ITEM_WIDTH/2;
      bp->y = y - ITEM_HEIGHT/2;  
      
      XFillRectangle(xdata.display, bp->pm, xdata.RevGC, 
                     0, 0, ITEM_WIDTH, ITEM_HEIGHT);
      XCopyArea(xdata.display,
   	        pixmaps[ball_i%MyMaxColor],
   	        bp->pm,
   	        xdata.ColorGC,
   	        0, 0, ITEM_WIDTH, ITEM_HEIGHT, 0, 0);
   }
} 
/*}}}  */

void
draw_pix(pix *bp)
/*{{{  */
{
   if ((bp->aw != 0) && (bp->w !=0))
      XCopyArea(xdata.display, bp->pm, xdata.drawWin, xdata.XorGC,
   	        0, 0, bp->w, bp->h,
   	        bp->x, bp->y);  
} 
/*}}}  */

void
undraw_pix(pix *bp)
/*{{{  */
{
   draw_pix(bp);
} 
/*}}}  */

void
acid_draw_pix(pix *bp)
/*{{{  */
{
   if ((bp->aw != 0) && (bp->w !=0))
      XCopyArea(xdata.display, bp->pm, xdata.drawWin, xdata.ColorGC,
   	        0, 0, bp->w, bp->h,
   	        bp->x, bp->y);  
} 
/*}}}  */

