/*{{{  Header*/
/*
 * Basic X program framework from xball:
 * Drawing Balls, Handling X Events, and Stuff.
 *
 * Put together by : Joe Gaudreau.
 *
 * Thanx to David Nedde (daven@maxine.wpi.edu) for writing Xball.
 *
 */

/* Changed 16/9/91 by Alastair Reid to allow resizing.
 *   Principal change is that xview now deals with a "virtual"
 *   interface STD_WIDTH wide and STD_HEIGHT high and that
 *   xTens magnifies any coordinates according to the current
 *   magnification as determined by current window size.
 * Changed 17/9/91 by Alastair Reid to return keyboard input to
 *   main program.
 */
/*}}}  */
/*{{{  Includes*/
#include <stdio.h>
#include <string.h>

#define NeedFunctionPrototypes 1

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeBG.h>
#include <Xm/ScrollBar.h>
#include <Xm/ToggleB.h>
#include <Xm/SeparatoG.h>
#include <Xm/SelectioB.h>
#include <Xm/List.h>
#include <Xm/LabelG.h>
#include <Xm/ArrowBG.h>
#include <Xm/DialogS.h>

/*
#include "widgets/Hand.h"
*/

#include "3d.h"
#include "animate.h"   /* Do we want this ? */
#include "juggle.h"
#include "siteswap.h"
#include "globals.h"

#include "x_uif.h"

#include "data/play.bmp"
#include "data/pause.bmp"
#include "data/rewind.bmp"
#include "data/reset.bmp"
#include "data/warpOn.bmp"
#include "data/warpOff.bmp"
/*}}}  */
/*{{{  Defines*/
extern void redo_screen(void);

#ifdef DEBUG
#define debugf(x) printf(x)
#else
#define debugf(x)
#endif

#define TRUE  1
#define FALSE 0

#define MAX_DIM (unsigned int)(1<<(8*sizeof(Position))) - 1

/*{{{  Coordinate space conversion macros*/
#define X_TO_X(x) \
( \
   ((x-STD_WIDTH/2)  * cur_width)  / STD_WIDTH  + cur_width/2 \
)
#define Y_TO_X(y) \
( \
   ((y-STD_HEIGHT/2) * cur_height) / STD_HEIGHT + cur_height/2 \
)

#define X_TO_J(x) \
( \
   (x * STD_WIDTH) / cur_width \
)

#define Y_TO_J(y) \
( \
   (y * STD_HEIGHT) / cur_height \
)
/*}}}  */

#define D_LEFT    0
#define D_RIGHT   1
#define D_UP      2
#define D_DOWN    3
#define D_IN      4
#define D_OUT     5
#define D_NONE    -1
#define D_RESET   -2
/*}}}  */
/*{{{  Globals*/

/* Look ma, them thar's global variables */

/* 
 * Mmm, is this structure thing really a good plan? 
 */
XDataType xdata;

XEvent    MyWinEvent;
KeySym    MyKey;
Widget    hand_w[MAX_NUM_HANDS];
Widget    objectPopup, msg_win;

Pixmap    reset_pix, warpOn_pix, warpOff_pix;

siteswap   **history = NULL;
int        hist_pos = 0,
           hist_len = 0;

double     a=0.0,
           b=0.0;
int        view_delta   = D_NONE,
           hand_delta   = D_NONE;

int        step = 20;

boolean    perspective = TRUE;
boolean    rotational  = TRUE,
           move_coi    = FALSE;
short      clubs       = TRUE, 
           fill        = TRUE;

int        cur_hand    = 0,
           cur_catch   = 0;

int        cur_width  = INIT_WIDTH, 
           cur_height = INIT_HEIGHT;

int de,       /* Distance of 'screen' from eye */
    ds,       /* Position of screen (from origin) */
    clip_delta = 50;


tx         view_tx;
point      eye, coi;

Polyinfo   props[] = {
#include "data/objclubE.h"
#include "data/objclubA.h"
#include "data/objknife.h"
#include "data/objring.h"
#include "data/objico.h"
#include "data/objocta.h"
#include "data/objpyr.h"
};

int   props_size  = sizeof(props)/sizeof(props[0]);
int   prop_index  = 0;

extern double sqrt(double);
extern double sincos(double,double*,double*);
/*}}}  */

void
show_msg(char *msg, int is_error)
/*{{{  */
{
   if (msg_win == NULL)
   {
      printf("%s",msg);
      return;
   }
   
   XtVaSetValues(msg_win,
                 XmNdialogType,    XmDIALOG_ERROR,
                 XmNmessageString, 
                    /* Use this to deal with \ns */
                    XmStringCreateLtoR(msg, XmSTRING_DEFAULT_CHARSET),
                 NULL);
   
   if (is_error)
   {
      XtVaSetValues(msg_win, 
         XmNdialogTitle,   XmStringCreateSimple("Site-swap Error"), 
         XmNokLabelString, XmStringCreateSimple("  Arse !  "),
         NULL);
   }
   else
   {
      XtVaSetValues(msg_win, 
         XmNdialogTitle,   XmStringCreateSimple("Site-swap Information"), 
         XmNokLabelString, XmStringCreateSimple("  Heh, cool!  "),
         NULL);
   }
   
   XtManageChild(msg_win);
}
/*}}}  */

void
fit_view_to_pattern()
/*{{{  */
{
   int h;
   int max_hand_y = 0, 
       min_dwell  = 4;
   int max_T, max_S;
      
   /* 
    * Calculate the height of the highest throw in the pattern
    */
   for(h=0; h<cur_ss->hands; h++)
   {
      max_hand_y = max(max_hand_y, max(hands[h].catch.y,hands[h].throw.y));
      min_dwell  = min(min_dwell, hands[h].dwell);
   }
                 
   max_T = (cur_ss->highest*2 - min_dwell)*cur_ss->height;
   max_S = (max_T*max_T/4) >>3 ;
   
   coi.x = STD_WIDTH2;
   coi.y = (max_hand_y + max_S)/2;
   coi.z = 0;
   
   de = -500;
   ds = 1000 + max(INIT_DS, -de*( (max_S - max_hand_y)/STD_HEIGHT + 1) );
      
   identity_tx(&view_tx);
   rotate_tx(&view_tx, 'y', 9*PI/8);
   pre_offset_tx(&view_tx, &coi);
}
/*}}}  */

int
highly_intelligent_error_handler(Display *d, XErrorEvent *e)
/*{{{  */
{
   printf("\n\n\n******** Oh bugger! ********\n\n\n");
}
/*}}}  */

void 
init_graphics(int *argcp, char *argv[])
/*{{{  */
{
   String    fallbacks[] = {"*XmDrawingArea.background: white",
                            "*title: xj - Virtual Juggling", 
                            "*keyboardFocusPolicy: POINTER",
                            NULL};
   
   xdata.topWidget = XtAppInitialize(&xdata.appContext, "MJuggle", NULL,0,
                                     argcp, argv, fallbacks, NULL, 0);

   xdata.display = XtDisplay(xdata.topWidget);
   xdata.screen  = DefaultScreen(xdata.display);
}  
/*}}}  */

void
add_ss(siteswap *ss, XmString s)
/*{{{  */
{
   if (history == NULL)
   {
      hist_len = 10;
      hist_pos = 0;
      history = (siteswap**)malloc(hist_len*sizeof(siteswap*));
   }
   
   if (hist_pos == hist_len)
   {
      hist_len *= 2;
      history = (siteswap**)realloc(hist_len*sizeof(siteswap*));
   }
   
   history[hist_pos++] = ss;
   
   XmListAddItem(xdata.patternList, s, hist_pos);
}
/*}}}  */

/*{{{  Callback routines*/
void
exitCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   pausing = 0;
   done = TRUE;
}
/*}}}  */

void
checkWindowSize()
/*{{{  */
{
   Dimension wd, ht, b;
   Arg       al[] = { {XmNwidth,       (XtArgVal)&wd},
                      {XmNheight,      (XtArgVal)&ht},
                      {XmNborderWidth, (XtArgVal)&b}};

   XtGetValues(xdata.drawArea, al, XtNumber(al));
   
   /*printf("Resize (%d,%d) -> (%d,%d) Border %d\n",
            cur_width, cur_height, wd, ht, b);*/
   
   cur_width  = wd;
   cur_height = ht;
}
/*}}}  */

void
drawCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   checkWindowSize();
   redo_screen();
}
/*}}}  */

void
resizeCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   checkWindowSize();
   
   if (xdata.drawWin != NULL)
      redo_screen();

   debugf(("Done Resize\n");)
}
/*}}}  */
 
void
detailCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   cur_ss->detail = ((XmScrollBarCallbackStruct*)call)->value;
   debugf(("Detail = %d\n",cur_ss->detail);)
}
/*}}}  */

void
heightCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   cur_ss->height = ((XmScrollBarCallbackStruct*)call)->value;
   debugf(("Height = %d\n",cur_ss->height);)
}
/*}}}  */

void
scoopCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
/*   cur_ss->scoop = ((XmScrollBarCallbackStruct*)call)->value;*/
}
/*}}}  */

void
toggleButtonCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   boolean set    = ((XmToggleButtonCallbackStruct*)call)->set;
   *(boolean*)cli = set;
   
   /*
    * Horrible Hack Alert
    */
   if ( (boolean*)cli == &perspective)
   {
      ds += 2*de;
      de = -de;
   }
      
   redo_screen();
}
/*}}}  */

void
popupCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   XtManageChild((Widget)cli);
}
/*}}}  */

void
popdownCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   XtUnmanageChild((Widget)cli);
}
/*}}}  */


/*{{{  Video buttons*/
void
playButtonCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   next_pause = 0;
   pausing = 0;
}
/*}}}  */

void
stopButtonCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   pausing = 1;
}
/*}}}  */

void
pauseButtonCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   next_pause = 1;
   pausing = 0;
}
/*}}}  */

void
rewindButtonCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   new_ss = NULL;
   pausing = 0;
}
/*}}}  */
/*}}}  */

/*{{{  Pattern selection*/
void
newPatternCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   XmSelectionBoxCallbackStruct *s = (XmSelectionBoxCallbackStruct*)call;
   siteswap                     *ss = (siteswap*)malloc(sizeof(siteswap));
   char                         *text;
   int                          pos;
   
   if (pos = XmListItemPos(xdata.patternList, s->value))
   {
      new_ss = history[pos-1];
   
      XtVaSetValues(xdata.current,
                    XmNlabelString, s->value,
                    NULL);
   }
   else
   {
      XmStringGetLtoR(s->value, XmSTRING_DEFAULT_CHARSET, &text);
      strcpy(ss->pattern_str, text);
      XtFree(text);
      
      ss->startup_str[0] = '\0';
      ss->startup_len = 0;
      ss->hands = nhands;      /* Keep same number as startup */
      
      extract_pattern(ss);
      if (ss->valid)
      {
         new_ss = ss;                           /* Inform main loop of change */
         add_ss(ss, s->value);                  /* Add to history */
         XtVaSetValues(xdata.current,           /* Set current title string */
                       XmNlabelString, s->value,
                       NULL);
      }
      else
         free(ss);
   }         
}     
/*}}}  */

void
oldPatternCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   XmString s   = ((XmSelectionBoxCallbackStruct*)call)->value;
   int      pos = XmListItemPos(xdata.patternList, s);
   
   new_ss = history[pos-1];
   
   XtVaSetValues(xdata.current,
                 XmNlabelString, s,
                 NULL);
}
/*}}}  */
/*}}}  */

/*{{{  Object selection*/
void
popObjectCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   XtManageChild(objectPopup);
}
/*}}}  */

void
changeObjectCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   prop_index = (int)cli;
   
   redo_screen();
}
/*}}}  */
/*}}}  */

/*{{{  Fly buttons*/
void
flyButtonArmCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int dir = (int)cli;
   
   view_delta = dir;
   
   if ( !move_coi )
      switch (dir)
      {
      case D_RIGHT : rotate_tx(&view_tx, 'y', 0.1);  break;
      case D_LEFT  : rotate_tx(&view_tx, 'y', -0.1); break;
      case D_UP    : rotate_tx(&view_tx, 'x', 0.1);  break;
      case D_DOWN  : rotate_tx(&view_tx, 'x', -0.1); break;
      case D_OUT   : ds+=2*step; break;
      case D_IN    : ds-=2*step; 
                     if ( ds + de < 0 ) 
                     { 
                        rotate_tx(&view_tx, 'y', PI);
                        view_delta = D_NONE;
                     } 
                     break;
      }
   else
   {
      point d = {0,0,0};
      
      switch (dir)
      {              
      case D_RIGHT : d.x = step;  break;
      case D_LEFT  : d.x = -step; break;
      case D_UP    : d.y = step;  break;
      case D_DOWN  : d.y = -step; break;
      case D_OUT   : d.z = step;  break;
      case D_IN    : d.z = -step; break;
      }
   
      tx_point(&d, &view_tx, &d);
      d.x -= view_tx.dx;
      d.y -= view_tx.dy;
      d.z -= view_tx.dz;
      add_point(&coi, &d, &coi);
      
      pre_offset_tx(&view_tx, &coi);
   }
   
   /*{{{  Old version*/
   /*
      printf("Before: Eye pos (%d,%d,%d) a=%.1lf b=%.1lf ds=%d de=%d\n",
               eye.x,eye.y,eye.z,
               a,b,ds,de);
               
   if (rotational)
   {
      double sina,cosa,sinb,cosb;
      
      sincos(a, &sina, &cosa);
      sincos(b, &sinb, &cosb);
      
      eye.x = sina * cosb * ds;
      eye.y = sinb * ds;
      eye.z = cosa * cosb * ds;
   }
   else
   {
      int l2 = eye.x*eye.x + eye.z*eye.z;
      
      a  = atan( (double)eye.z/eye.x );
      b  = atan( (double)eye.y/sqrt((double)l2) );
      ds = sqrt( (double)l2 + (double)eye.y*eye.y );
   }
      printf("After: Eye pos (%d,%d,%d) a=%.1lf b=%.1lf ds=%d de=%d\n",
               eye.x,eye.y,eye.z,
               a,b,ds,de);
   
   set_view_tx();
   */
   /*}}}  */
   
   redo_screen();
   
   debugf(("a=%.3f b = %.3f Eye Z = %d\n",a,b,ds+de);)

}
/*}}}  */

void
flyButtonDisarmCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   view_delta = D_NONE;

   if ( (int)cli == D_RESET )
   {
      fit_view_to_pattern();
      redo_screen();
   }
}
/*}}}  */
/*}}}  */

/*{{{  Hand related buttons*/
void
popupHCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int h;
   
   if (cur_hand < 0)
      for(h=0; h<nhands; h++)
         display_point[h+1] = cur_catch ? &hands[cur_hand].throw : 
                                          &hands[cur_hand].catch;
   else
      display_point[cur_hand+1] = cur_catch ? &hands[cur_hand].throw : 
                                              &hands[cur_hand].catch;
   
   redo_screen();
   
   XtManageChild((Widget)cli);
}
/*}}}  */

void
popdownHCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int h;
   
   if (cur_hand < 0)
      for(h=0; h<nhands; h++)
         display_point[h+1] = NULL;
   else
      display_point[cur_hand+1] = NULL;
                                     
   redo_screen();
   XtUnmanageChild((Widget)cli);
}
/*}}}  */

void
step_point(point *p, int dir)
/*{{{  */
{   
   switch (dir)
   {              
   case D_RIGHT : p->x += step;  break;
   case D_LEFT  : p->x += -step; break;
   case D_UP    : p->y += step;  break;
   case D_DOWN  : p->y += -step; break;
   case D_OUT   : p->z += step;  break;
   case D_IN    : p->z += -step; break;
   }
}
/*}}}  */

void
handPosButtonArmCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int dir = (int)cli;
   int h;
   
   hand_delta = dir;
      
   if (cur_hand < 0)
      for(h=0; h<nhands; h++)
         switch(cur_catch)
         {
         case 0:  step_point(&hands[h].catch, dir); break;
         case 1:  step_point(&hands[h].throw, dir); break;
         case -1: step_point(&hands[h].throw, dir); 
                  step_point(&hands[h].catch, dir); break;
         }
   else
      switch(cur_catch)
      {
      case 0:  step_point(&hands[cur_hand].catch, dir); break;
      case 1:  step_point(&hands[cur_hand].throw, dir); break;
      case -1: step_point(&hands[cur_hand].throw, dir); 
               step_point(&hands[cur_hand].catch, dir); break;
      }
   
   redo_screen();
}
/*}}}  */

void
handPosButtonDisarmCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int h;
   
   hand_delta = D_NONE;

   if ((int)cli == D_RESET)
   {
      if (cur_hand < 0)
            for(h=0; h<nhands; h++)
               switch(cur_catch)
               {
               case 0:  hands[h].catch = hands[h].init_catch; break;
               case 1:  hands[h].throw = hands[h].init_throw; break;
               case -1: hands[h].throw = hands[h].init_throw;
                        hands[h].catch = hands[h].init_catch; break;
               }
         else
            switch(cur_catch)
            {
            case 0:  hands[cur_hand].catch = hands[cur_hand].init_catch; break;
            case 1:  hands[cur_hand].throw = hands[cur_hand].init_throw; break;
            case -1: hands[cur_hand].throw = hands[cur_hand].init_throw;
                     hands[cur_hand].catch = hands[cur_hand].init_catch; break;
            }
      
      XtManageChild(XtParent(w));
      redo_screen();
   }
}
/*}}}  */

void
changeHandCB(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int new_hand = (int) cli;
   int h;
   
   if (new_hand == cur_hand)
      return;
   
   if (cur_hand < 0)
      for(h=0; h<nhands; h++)
         display_point[h+1] = NULL;
   else
      display_point[cur_hand+1] = NULL;
      
   if (new_hand < 0)
      for(h=0; h<nhands; h++)
         display_point[h+1] = cur_catch ? &hands[h].throw : 
                                          &hands[h].catch;
   else
      display_point[new_hand+1] = cur_catch ? &hands[new_hand].throw : 
                                              &hands[new_hand].catch;
      
   
   if (new_hand < 0)
      for(h=0; h<nhands; h++)
         hands[h].dwell = hands[cur_hand].dwell;
   else
      XtVaSetValues(xdata.dwell,
                    XmNvalue,   hands[new_hand].dwell,
                    NULL);
   
   debugf(("changeHandCB = cur_hand = %d, cur_catch = %d\n",
          cur_hand,cur_catch);)
   
   cur_hand = new_hand;

   redo_screen();
}
/*}}}  */

void
changeCatchCB(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int new_catch = (int)cli;
   int h;
   
   if (cur_hand < 0)
      for(h=0; h<nhands; h++)
         display_point[h+1] = new_catch ? &hands[h].throw : 
                                          &hands[h].catch;
   else
      display_point[cur_hand+1] = new_catch ? &hands[cur_hand].throw : 
                                              &hands[cur_hand].catch;
   
   cur_catch = new_catch;
   
   debugf(("changeCatchCB = cur_hand = %d, cur_catch = %d\n",
          cur_hand,cur_catch);)
   
   redo_screen();
}
/*}}}  */

void
dwellCallBack(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   int newVal = ((XmScrollBarCallbackStruct*)call)->value;
   int h;
   
   if (cur_hand < 0)
      for(h=0; h<nhands; h++)
         hands[h].dwell = newVal;
   else
      hands[cur_hand].dwell = newVal;
}
/*}}}  */

/*}}}  */

void
changeCOICB(Widget w, XtPointer cli, XtPointer call)
/*{{{  */
{
   move_coi = (int) cli;
   display_point[0] = move_coi ? &coi : NULL;
   redo_screen();
}
/*}}}  */


/*}}}  */

Widget
create_pattern_selector(Widget parent)
/*{{{  */
{
   Widget popup, select;
   
   popup = XmCreateDialogShell(parent, "patPopup", NULL, 0);
   XtVaSetValues(popup,
                 XmNtitle, "xj - Patterns",
                 NULL);
   
   select = XtVaCreateWidget("selector", 
                            xmSelectionBoxWidgetClass,
                            popup,
                            XmNlistVisibleItemCount, 7,
                            XmNmustMatch,            True,
                            NULL);
   
   XtUnmanageChild(XmSelectionBoxGetChild(select, XmDIALOG_LIST_LABEL));
   XtUnmanageChild(XmSelectionBoxGetChild(select, XmDIALOG_CANCEL_BUTTON));
   XtUnmanageChild(XmSelectionBoxGetChild(select, XmDIALOG_HELP_BUTTON));
   XtUnmanageChild(XmSelectionBoxGetChild(select, XmDIALOG_HELP_BUTTON));
   
   XtAddCallback(select, XmNnoMatchCallback, newPatternCallBack,  NULL);
   XtAddCallback(select, XmNokCallback,      oldPatternCallBack,  NULL);
   XtAddCallback(select, XmNapplyCallback,   newPatternCallBack,  NULL);

   return select;
}
/*}}}  */
   
Widget
create_hand_editor(Widget parent)
/*{{{  */
{
   Widget edit, buttonBar, aheadView, sideView, tmp;
   Widget handsB, handsMenu, catchB, catchMenu;
   Widget dwellLab, dwellBar;
   char   name[40];
   Arg    al[1];
   int    h;
   
   /*{{{  FormDialog */
   XtSetArg(al[0], XmNtitle, "xj - Hand Editor");
   
   edit = XmCreateFormDialog(parent, "handEdit", al, 1);
   /*}}}  */
   /*{{{  COMMENT MenuBar */
   /*
   buttonBar = XmCreateRowColumn(edit, "buttons", NULL, 0);
   XtManageChild(buttonBar);
   */
   /*}}}  */
   /*{{{  Hands menu*/
   handsMenu = XmCreatePulldownMenu(edit, "handsMenu", NULL, 0);
   
   for(h=0; h<nhands; h++)
   {
      sprintf(name,"Hand %d",h);
      tmp = XtVaCreateManagedWidget(name, xmPushButtonGadgetClass, handsMenu, 
                                    NULL);
      XtAddCallback(tmp, XmNactivateCallback, 
                       changeHandCB, (XtPointer)h);
   }
   
   tmp = XtVaCreateManagedWidget("All hands", 
                                 xmPushButtonGadgetClass, handsMenu, 
                                 NULL);
   XtAddCallback(tmp, XmNactivateCallback, 
                    changeHandCB, (XtPointer)-1);
   
   
   XtSetArg(al[0], XmNsubMenuId,    handsMenu);
   XtSetArg(al[1], XmNwhichButton,  Button1);
   handsB = XmCreateOptionMenu(edit, "hands", al, 2);
   XtManageChild(handsB);
   /*}}}  */
   /*{{{  Catch menu*/
   catchMenu = XmCreatePulldownMenu(edit, "catchMenu", NULL, 0);
   
   tmp = XtVaCreateManagedWidget("Catch Position", 
                                 xmPushButtonGadgetClass, catchMenu, 
                                 NULL);
   XtAddCallback(tmp, XmNactivateCallback, 
                    changeCatchCB, (XtPointer)0);
   tmp = XtVaCreateManagedWidget("Throw Position", 
                                 xmPushButtonGadgetClass, catchMenu, 
                                 NULL);
   XtAddCallback(tmp, XmNactivateCallback, 
                    changeCatchCB, (XtPointer)1);
   tmp = XtVaCreateManagedWidget("Both Positions", 
                                 xmPushButtonGadgetClass, catchMenu, 
                                 NULL);
   XtAddCallback(tmp, XmNactivateCallback, 
                    changeCatchCB, (XtPointer)-1);
                  
   XtSetArg(al[0], XmNsubMenuId, catchMenu);
   XtSetArg(al[1], XmNwhichButton,  Button1);
   catchB = XmCreateOptionMenu(edit, "catchs", al, 2);
   XtManageChild(catchB);
   XtVaSetValues(catchB,
                 XmNleftAttachment, XmATTACH_WIDGET,
                 XmNleftWidget,     handsB,
                 NULL);
   /*}}}  */

   /*{{{  Steer buttons*/
   tmp = XtVaCreateManagedWidget("reset", xmPushButtonGadgetClass,
                  edit, 
                  XmNleftAttachment, XmATTACH_FORM,
                  XmNleftOffset,     26+12, 
                  XmNtopAttachment,  XmATTACH_WIDGET, 
                  XmNtopWidget,      catchB, 
                  XmNtopOffset,      26+10,
                  XmNlabelType,      XmPIXMAP, 
                  XmNlabelPixmap,    reset_pix,
                  NULL);
   
   XtAddCallback(tmp, XmNactivateCallback, 
                 handPosButtonDisarmCallBack, (XtPointer)D_RESET);
   
   
   #define dirButton(name, motif_dir, my_dir, x, y) \
   tmp = XtVaCreateManagedWidget(name, xmArrowButtonGadgetClass, edit, \
                  XmNarrowDirection, motif_dir, \
                  XmNleftAttachment, XmATTACH_FORM, \
                  XmNleftOffset,     x+12, \
                  XmNtopAttachment,  XmATTACH_WIDGET, \
                  XmNtopWidget,      catchB, \
                  XmNtopOffset,      y+10, \
                  NULL); \
   XtAddCallback(tmp, XmNarmCallback, \
                 handPosButtonArmCallBack, (XtPointer)my_dir); \
   XtAddCallback(tmp, XmNdisarmCallback, \
                 handPosButtonDisarmCallBack, (XtPointer)my_dir)
   
   dirButton("in",   XmARROW_UP,    D_IN,    90, 10);
   dirButton("out",  XmARROW_DOWN,  D_OUT,   90, 46);
   dirButton("left", XmARROW_LEFT,  D_LEFT,   0, 28);
   dirButton("right",XmARROW_RIGHT, D_RIGHT, 56, 28);
   dirButton("up",   XmARROW_UP,    D_UP,    28,  0);
   dirButton("down", XmARROW_DOWN,  D_DOWN,  28, 56);
   #undef dirButton
   /*}}}  */

   /*{{{  Dwell slider*/
   dwellLab = XtVaCreateManagedWidget("Dwell : ", xmLabelGadgetClass,
                  edit,
                  XmNtopAttachment,    XmATTACH_WIDGET,
                  XmNtopWidget,        tmp,
                  XmNtopOffset,        4,
                  /*
                  XmNbottomAttachment, XmATTACH_FORM,
                  */
                  XmNbottomOffset,     4,
                  XmNleftAttachment,   XmATTACH_FORM,
                  XmNleftOffset,       4,
                  NULL);
                                     
   dwellBar = XtVaCreateManagedWidget("dwell", xmScrollBarWidgetClass, 
                  edit,
                  XmNorientation,      XmHORIZONTAL,
                  XmNminimum,          0,
                  XmNmaximum,          nhands*2+1,
                  XmNincrement,        1,
                  XmNpageIncrement,    1,
                  XmNvalue,            2,
                  XmNtopAttachment,    XmATTACH_WIDGET,
                  XmNtopWidget,        tmp,
                  XmNtopOffset,        4,
                  /*
                  XmNbottomAttachment, XmATTACH_FORM,
                  */
                  XmNbottomOffset,     4,
                  XmNleftAttachment,   XmATTACH_WIDGET,
                  XmNleftWidget,       dwellLab,
                  XmNrightAttachment,  XmATTACH_FORM,
                  XmNrightOffset,      4,
                  NULL);
   
   XtAddCallback(dwellBar, XmNvalueChangedCallback, dwellCallBack, NULL);
   
   xdata.dwell = dwellBar;
   /*}}}  */
   
   /*{{{  Dismiss*/
   tmp = XtVaCreateManagedWidget("Dismiss", xmPushButtonGadgetClass, edit, 
                                 XmNrightAttachment,  XmATTACH_FORM,
                                 XmNrightOffset,      8,
                                 XmNbottomAttachment, XmATTACH_WIDGET,
                                 XmNbottomWidget,     dwellBar,
                                 XmNbottomOffset,     8,
                                 NULL);
                                 
   XtAddCallback(tmp, XmNactivateCallback, 
                 popdownHCallBack, (XtPointer)edit);
   /*}}}  */
   
   return edit;
}
/*}}}  */

void
setup_uif(void)
/*{{{  */
{
   Display  *disp   = xdata.display;
   int      screen  = xdata.screen;
   Window   rootWin = DefaultRootWindow(disp);
   int      depth   = DefaultDepth(disp,screen);
   Widget   mainWin, menuBar, form, drawForm, drawArea;
   Widget   fileMenu, fileButton, exitButton, helpButton;
   Widget   controls, stopB, playB, pauseB, rewindB;
   Widget   detailLab, detailBar;
   Widget   heightLab, heightBar;
   Widget   patternButton, patternSelect, objectButton, current;
   Widget   flyBox, upB, downB, leftB, rightB, inB, outB;
   Widget   tmpWidget, sep1, sep2, sep3;
   Widget   submenu, optionsButton, displayButton;
   Widget   handsButton;
   Widget   warpLab, warpLight;
   Widget   coiB, coiMenu;
   Arg      al[1];
   XmString labelStr;
   int      h,p;
   Pixmap   px;
   Pixel    fgPixel, bgPixel;
   
   /*{{{  Debugging*/
   if (debug_x)
   {
      XSetErrorHandler( highly_intelligent_error_handler );
      XSynchronize(xdata.display, True);
   }
   /*}}}  */
   
   map_colours(&fgPixel, &bgPixel);
   
   /*{{{  Pixmaps */
   reset_pix = XCreatePixmapFromBitmapData(disp, rootWin, 
                                           reset_bits, 
                                           reset_width, 
                                           reset_height,
                                           fgPixel,
                                           bgPixel,
                                           depth),
   warpOn_pix = XCreatePixmapFromBitmapData(disp, rootWin, 
                                           warpOn_bits, 
                                           warpOn_width, 
                                           warpOn_height,
                                           fgPixel,
                                           bgPixel,
                                           depth),
   warpOff_pix = XCreatePixmapFromBitmapData(disp, rootWin, 
                                           warpOff_bits, 
                                           warpOff_width, 
                                           warpOff_height,
                                           fgPixel,
                                           bgPixel,
                                           depth),
   /*}}}  */
   /*{{{  Main window*/
   mainWin = XtVaCreateManagedWidget("main win", xmMainWindowWidgetClass,
               xdata.topWidget,
               NULL);
   /*}}}  */

   /*{{{  Message window*/
   msg_win = (Widget)XmCreateMessageDialog(mainWin, "msgs", NULL, 0);
   XtUnmanageChild( (Widget)XmMessageBoxGetChild(msg_win, XmDIALOG_HELP_BUTTON));
   XtUnmanageChild( (Widget)XmMessageBoxGetChild(msg_win, XmDIALOG_CANCEL_BUTTON));
   /*}}}  */
   
   /*{{{  Menu bar*/
   menuBar = XmCreateMenuBar(mainWin, "menu bar", NULL, 0);
   XtManageChild(menuBar);
   /*}}}  */
   /*{{{  File menu*/
   fileButton = XtVaCreateManagedWidget("File", xmCascadeButtonGadgetClass,
                                         menuBar, NULL);
   fileMenu   = XmCreatePulldownMenu(menuBar, "file menu", NULL, 0);
   
   XtVaSetValues(fileButton, XmNsubMenuId, fileMenu, NULL);
   
   exitButton = XtCreateManagedWidget("Exit", xmPushButtonGadgetClass, 
                  fileMenu, NULL, 0);
   XtAddCallback(exitButton, XmNactivateCallback, exitCallBack, NULL);
   /*}}}  */
   /*{{{  Options menu*/
   optionsButton = XtCreateManagedWidget("Options", 
                                         xmCascadeButtonGadgetClass,
                                         menuBar, NULL, 0);
   submenu = XmCreatePulldownMenu(menuBar, "submenu", NULL, 0);
   XtVaSetValues(optionsButton, XmNsubMenuId, submenu, NULL);
   
   #define toggleButton(name, toggle) \
   tmpWidget = XtVaCreateManagedWidget(name, xmToggleButtonWidgetClass, submenu, \
                                    XmNset,             toggle, \
                                    XmNborderWidth,     0, \
                                    NULL); \
   XtAddCallback(tmpWidget, XmNvalueChangedCallback, \
                 toggleButtonCallBack, &toggle)
   
   toggleButton("Show Grid",  show_grid);
   toggleButton("Show Cube",  show_cube);
   toggleButton("Show Arms",  show_arms);
   toggleButton("Show Robot", show_robot);
   
   #undef toggleButton
   /*}}}  */
   /*{{{  Toggles*/
   displayButton = XtCreateManagedWidget("Display", 
                                         xmCascadeButtonGadgetClass,
                                         menuBar, NULL, 0);
   submenu = XmCreatePulldownMenu(menuBar, "submenu2", NULL, 0);
   XtVaSetValues(displayButton, XmNsubMenuId, submenu, NULL);
   
   
   #define toggleButton(name, toggle) \
   tmpWidget = XtVaCreateManagedWidget(name, xmToggleButtonWidgetClass, submenu, \
                                    XmNset,             toggle, \
                                    XmNborderWidth,     0, \
                                    NULL); \
   XtAddCallback(tmpWidget, XmNvalueChangedCallback, \
                 toggleButtonCallBack, &toggle)
      
   toggleButton("Aciid!",      onacid);
   toggleButton("Perspective", perspective);
   toggleButton("Fill",        fill);
   toggleButton("Clubs",       clubs);
   
   #undef toggleButton
   /*}}}  */

   /*{{{  Objects menu*/
   objectButton = XtVaCreateManagedWidget("Objects", 
                            xmCascadeButtonGadgetClass,
                            menuBar,
                            NULL);
   
   objectPopup = XmCreatePulldownMenu(menuBar, "objPopup", NULL, 0);
   
   XtVaSetValues(objectButton, XmNsubMenuId, objectPopup, NULL);
   
   /*{{{  XmList vers*/
   /*
   XtVaSetValues(objectPopup,
                 XmNtitle, "Objects",
                 NULL);
   
   objectSelect = XtVaCreateWidget("objects", 
                                xmListWidgetClass,
                                objectPopup,
                                XmNvisibleItemCount, 7,
                                XmNselectionPolicy, XmSINGLE_SELECT,
                                XmNlistSizePolicy,  XmRESIZE_IF_POSSIBLE,
                                XmNlistLabelString, XmStringCreateSimple("Objects"),
                                NULL);
   XtAddCallback(objectSelect, XmNsingleSelectionCallback, changeObjectCallBack, NULL);
   */
   /*}}}  */
   
   for(p=0; p<props_size; p++)
   {
      tmpWidget = XtVaCreateManagedWidget(props[p].longname, 
                             xmPushButtonGadgetClass, 
                             objectPopup,
                             NULL);
      XtAddCallback(tmpWidget, XmNactivateCallback, 
                      changeObjectCallBack, (XtPointer)p);
   }
   
   tmpWidget = XtVaCreateManagedWidget("Mixture",
                          xmPushButtonGadgetClass, 
                          objectPopup,
                          NULL);
   XtAddCallback(tmpWidget, XmNactivateCallback, 
                   changeObjectCallBack, (XtPointer)-1);
   
   /*}}}  */
   
   /*{{{  COMMENT Help button*/
   /*
    * I suppose something ought to be put here ...
    *
    
   helpButton = XtVaCreateManagedWidget("Help", xmCascadeButtonGadgetClass,
                                         menuBar, NULL); 
   
   XtVaSetValues(menuBar, XmNmenuHelpWidget, helpButton, NULL);
   */
   /*}}}  */
    
   /*{{{  Form*/
   form = XtCreateManagedWidget("form",xmFormWidgetClass,
                                mainWin, NULL,0);
   /*}}}  */
   /*{{{  Current*/
   labelStr = XmStringCreate(cur_ss->pattern_str, XmSTRING_DEFAULT_CHARSET);
   
   current = XtVaCreateManagedWidget("current",
                                     xmLabelGadgetClass,
                                     form,
                                     XmNlabelString, labelStr,
                                     XmNalignment, XmALIGNMENT_CENTER,
                                     XmNborderWidth, 2,
                                     NULL);
   xdata.current = current;
   /*}}}  */
   /*{{{  Draw area*/
   drawArea = XtVaCreateManagedWidget("draw",xmDrawingAreaWidgetClass,
               form, 
               XmNwidth,  INIT_WIDTH,
               XmNheight, INIT_HEIGHT,
               NULL);   
               
   xdata.drawArea = drawArea;
   xdata.drawWin  = NULL; /* Wait till it's been realized */
   
   XtAddCallback(drawArea, XmNexposeCallback, drawCallBack, NULL);
   XtAddCallback(drawArea, XmNresizeCallback, resizeCallBack, NULL);
   /*}}}  */

   /*{{{  Control Panel*/
   controls = XtVaCreateManagedWidget("controls", xmFormWidgetClass, form,
                     NULL);
   
   #define SEP 8
   
   /*{{{  Video buttons*/
   #define videoButton(name) \
   name##B = XtVaCreateManagedWidget(#name, xmPushButtonGadgetClass, \
            controls, \
            XmNtopAttachment, XmATTACH_FORM, \
            XmNtopOffset,     SEP, \
            XmNlabelType,   XmPIXMAP, \
            XmNlabelPixmap, XCreatePixmapFromBitmapData(disp, rootWin, \
                                                  name##_bits, \
                                                  name##_width, \
                                                  name##_height, \
                                                  fgPixel, \
                                                  bgPixel, \
                                                  depth), \
            NULL); \
   XtAddCallback(name##B, XmNactivateCallback, name##ButtonCallBack, NULL)
   
   videoButton(play);
   videoButton(pause);
   videoButton(rewind);
   
   XtVaSetValues(playB,
                 XmNleftAttachment, XmATTACH_FORM,
                 XmNleftOffset,     SEP,
                 NULL);
   XtVaSetValues(pauseB,
                 XmNleftAttachment, XmATTACH_WIDGET,
                 XmNleftWidget,     playB,
                 NULL);
   XtVaSetValues(rewindB,
                 XmNleftAttachment, XmATTACH_WIDGET,
                 XmNleftWidget,     pauseB,
                 NULL);
   
   #undef videoButton
   /*}}}  */
   /*{{{  Pattern Selection */
   patternButton = XtVaCreateManagedWidget("Patterns ...", 
                            xmPushButtonGadgetClass,
                            controls,
                            XmNleftAttachment,   XmATTACH_FORM,
                            XmNleftOffset,       SEP,
                            XmNtopAttachment,    XmATTACH_WIDGET,
                            XmNtopWidget,        playB,
                            XmNtopOffset,        12,
                            NULL);
   
   patternSelect = create_pattern_selector(controls);
   
   XtAddCallback(patternButton, 
                 XmNactivateCallback, 
                 popupCallBack, 
                 (XtPointer)patternSelect);
   
   /*}}}  */
   /*{{{  Hand editor*/
   handsButton = XtVaCreateManagedWidget("Hands    ...", 
                            xmPushButtonGadgetClass,
                            controls,
                            XmNleftAttachment,   XmATTACH_FORM,
                            XmNleftOffset,       SEP,
                            XmNtopAttachment,    XmATTACH_WIDGET,
                            XmNtopWidget,        patternButton,
                            XmNtopOffset,        6,
                            NULL);
   
   XtAddCallback(handsButton, 
                 XmNactivateCallback, 
                 popupHCallBack, 
                 (XtPointer)create_hand_editor(controls));
   
   /*}}}  */
   
   /*{{{   Separator 1*/
   sep1 = XtVaCreateManagedWidget("sep1", xmSeparatorGadgetClass, controls,
                 XmNorientation,       XmVERTICAL,
                 XmNleftAttachment,    XmATTACH_WIDGET,
                 XmNleftWidget,        rewindB,
                 XmNleftOffset,        SEP,
                 XmNtopAttachment,     XmATTACH_FORM,
                 XmNtopOffset,         SEP,
                 XmNbottomAttachment,  XmATTACH_FORM,
                 XmNbottomOffset,      SEP,
                 NULL);
   /*}}}  */
   /*{{{  Flying Buttons*/
   /*{{{  COI/View menu*/
   coiMenu = XmCreatePulldownMenu(controls, "coiMenu", NULL, 0);
   
   tmpWidget = XtVaCreateManagedWidget("View Angle", 
                                 xmPushButtonGadgetClass, coiMenu, 
                                 NULL);
   XtAddCallback(tmpWidget, XmNactivateCallback, 
                    changeCOICB, (XtPointer)0);
   tmpWidget = XtVaCreateManagedWidget("Centre of Interest", 
                                 xmPushButtonGadgetClass, coiMenu, 
                                 NULL);
   XtAddCallback(tmpWidget, XmNactivateCallback, 
                    changeCOICB, (XtPointer)1);
                  
   XtSetArg(al[0], XmNsubMenuId, coiMenu);
   XtSetArg(al[1], XmNwhichButton,  Button1);
   coiB = XmCreateOptionMenu(controls, "coiB", al, 2);
   XtManageChild(coiB);
   XtVaSetValues(coiB,
                 XmNleftAttachment, XmATTACH_WIDGET,
                 XmNleftWidget,     sep1,
                 XmNleftOffset,     SEP,
                 XmNtopAttachment,  XmATTACH_FORM,
                 XmNtopOffset,      SEP,
                 NULL);
   /*}}}  */
   
   tmpWidget = XtVaCreateManagedWidget("centre", xmPushButtonGadgetClass, 
                  controls, 
                  XmNleftAttachment, XmATTACH_WIDGET,
                  XmNleftWidget,     sep1, 
                  XmNleftOffset,     26+SEP, 
                  XmNtopAttachment,  XmATTACH_WIDGET,
                  XmNtopWidget,      coiB,
                  XmNtopOffset,      26+SEP,
                  XmNlabelType,      XmPIXMAP, 
                  XmNlabelPixmap,    reset_pix,
                  NULL);
   
   XtAddCallback(tmpWidget, XmNactivateCallback, 
                 flyButtonDisarmCallBack, (XtPointer)D_RESET);
   
   #define dirButton(name, motif_dir, my_dir, x, y) \
   tmpWidget = XtVaCreateManagedWidget(name, xmArrowButtonGadgetClass, controls, \
                  XmNarrowDirection, motif_dir, \
                  XmNleftAttachment, XmATTACH_WIDGET, \
                  XmNleftWidget,     sep1, \
                  XmNleftOffset,     x+SEP, \
                  XmNtopAttachment,  XmATTACH_WIDGET, \
                  XmNtopWidget,      coiB, \
                  XmNtopOffset,      y+SEP, \
                  NULL); \
   XtAddCallback(tmpWidget, XmNarmCallback, \
                 flyButtonArmCallBack, (XtPointer)my_dir); \
   XtAddCallback(tmpWidget, XmNdisarmCallback, \
                 flyButtonDisarmCallBack, (XtPointer)my_dir)
   
   dirButton("left", XmARROW_LEFT,  D_LEFT,   0, 28);
   dirButton("right",XmARROW_RIGHT, D_RIGHT, 56, 28);
   dirButton("up",   XmARROW_UP,    D_UP,    28,  0);
   dirButton("down", XmARROW_DOWN,  D_DOWN,  28, 56);
   dirButton("in",   XmARROW_UP,    D_IN,    90, 10);
   dirButton("out",  XmARROW_DOWN,  D_OUT,   90, 46);
   /*}}}  */
   
   /*{{{   COMMENT Separator 2*/
   /*
   sep2 = XtVaCreateManagedWidget("sep2", xmSeparatorGadgetClass, controls,
               XmNorientation,       XmVERTICAL,
               XmNleftAttachment,    XmATTACH_WIDGET,
               XmNleftWidget,        coiB,
               XmNleftOffset,        4,
               XmNtopAttachment,     XmATTACH_FORM,
               XmNtopOffset,         4,
               XmNbottomAttachment,  XmATTACH_FORM,
               XmNbottomOffset,      4,
               NULL);
   */
   /*}}}  */
   
   /*{{{   Separator 3*/
   sep3 = XtVaCreateManagedWidget("sep3", xmSeparatorGadgetClass, controls,
               XmNorientation,       XmVERTICAL,
               XmNleftAttachment,    XmATTACH_WIDGET,
               XmNleftWidget,        coiB,
               XmNleftOffset,        SEP,
               XmNtopAttachment,     XmATTACH_FORM,
               XmNtopOffset,         SEP,
               XmNbottomAttachment,  XmATTACH_FORM,
               XmNbottomOffset,      SEP,
               NULL);
   /*}}}  */
   /*{{{  Height slider*/
   heightLab = XtVaCreateManagedWidget("Height: ", xmLabelGadgetClass,
                  controls,
                  XmNtopAttachment,   XmATTACH_FORM,
                  XmNtopOffset,       SEP,
                  XmNleftAttachment,  XmATTACH_WIDGET,
                  XmNleftWidget,      sep3,
                  XmNleftOffset,      SEP,
                  NULL);
                                     
   heightBar = XtVaCreateManagedWidget("height", xmScrollBarWidgetClass, 
                  controls,
                  XmNorientation,      XmHORIZONTAL,
                  XmNminimum,          1,
                  XmNmaximum,          200,
                  XmNincrement,        5,
                  XmNpageIncrement,    20,
                  XmNvalue,            cur_ss->height,
                  XmNtopAttachment,    XmATTACH_FORM,
                  XmNtopOffset,        SEP,
                  XmNbottomAttachment, XmATTACH_NONE,
                  XmNleftAttachment,   XmATTACH_WIDGET,
                  XmNleftWidget,       heightLab,
                  XmNrightAttachment,  XmATTACH_FORM,
                  XmNrightOffset,      SEP,
                  NULL);
   
   XtAddCallback(heightBar, XmNvalueChangedCallback, heightCallBack, NULL);
   
   xdata.height = heightBar;
   /*}}}  */
   /*{{{  Detail slider*/
   detailLab = XtVaCreateManagedWidget("Detail: ", xmLabelGadgetClass,
                  controls,
                  XmNtopAttachment,   XmATTACH_WIDGET,
                  XmNtopWidget,       heightLab,
                  XmNtopOffset,       SEP,
                  XmNleftAttachment,  XmATTACH_WIDGET,
                  XmNleftWidget,      sep3,
                  XmNleftOffset,      SEP,
                  NULL);
                                     
   detailBar = XtVaCreateManagedWidget("detail", xmScrollBarWidgetClass, 
                  controls,
                  XmNorientation,      XmHORIZONTAL,
                  XmNminimum,          2,
                  XmNmaximum,          50,
                  XmNincrement,        1,
                  XmNpageIncrement,    20,
                  XmNvalue,            cur_ss->detail,
                  XmNtopAttachment,    XmATTACH_WIDGET,
                  XmNtopWidget,        heightBar,
                  XmNtopOffset,        SEP,
                  XmNbottomAttachment, XmATTACH_NONE,
                  XmNleftAttachment,   XmATTACH_WIDGET,
                  XmNleftWidget,       detailLab,
                  XmNrightAttachment,  XmATTACH_FORM,
                  XmNrightOffset,      SEP,
                  NULL);
   
   XtAddCallback(detailBar, XmNvalueChangedCallback, detailCallBack, NULL);
   xdata.detail = detailBar;
   /*}}}  */
   /*{{{  Warp Light*/
   warpLab   = XtVaCreateManagedWidget("Reality Warp:", xmLabelGadgetClass,
                  controls,
                  XmNleftAttachment,    XmATTACH_WIDGET,
                  XmNleftWidget,        sep3,
                  XmNleftOffset,        SEP,
                  XmNtopAttachment,     XmATTACH_WIDGET,
                  XmNtopWidget,         detailLab,
                  XmNtopOffset,         SEP,
                  NULL);
   
   /* Note: created UNmanaged */
   warpLight = XtVaCreateWidget("warp", xmLabelGadgetClass,
                  controls,
                  XmNleftAttachment,    XmATTACH_WIDGET,
                  XmNleftWidget,        warpLab,
                  XmNleftOffset,        12,
                  XmNtopAttachment,     XmATTACH_WIDGET,
                  XmNtopWidget,         detailBar,
                  XmNtopOffset,         SEP,
                  XmNlabelType,         XmPIXMAP,
                  XmNlabelPixmap,            reset_pix,
                  NULL);
                  
   xdata.warp = warpLight;
   /*}}}  */
   
   /*}}}  */
   
   /*{{{  Form Attachements*/
   XtVaSetValues(current,
         XmNleftAttachment,   XmATTACH_FORM,
         XmNrightAttachment,  XmATTACH_FORM,
         XmNtopAttachment,    XmATTACH_FORM,
         XmNbottomAttachment, XmATTACH_NONE,
         NULL);
            
   XtVaSetValues(drawArea,
         XmNtopAttachment,    XmATTACH_WIDGET,
         XmNtopWidget,        current,
         XmNbottomAttachment, XmATTACH_WIDGET,
         XmNbottomWidget,     controls,
         XmNleftAttachment,   XmATTACH_FORM,
         XmNrightAttachment,  XmATTACH_FORM,
         XmNborderWidth,      2,
         XmNleftOffset,       5,
         XmNrightOffset,      5,
         XmNtopOffset,        5,
         XmNbottomOffset,     5,
         NULL);
   
   XtVaSetValues(controls,
         XmNtopAttachment,    XmATTACH_NONE,
         XmNbottomAttachment, XmATTACH_FORM,
         XmNleftAttachment,   XmATTACH_FORM,
         XmNleftWidget,       drawArea,
         XmNrightAttachment,  XmATTACH_FORM,
         XmNborderWidth,      2,
         XmNleftOffset,       5,
         XmNrightOffset,      5,
         XmNtopOffset,        5,
         XmNbottomOffset,     5,
         NULL);
   /*
   XtVaSetValues(patternSelect,
         XmNtopAttachment,    XmATTACH_FORM,
         XmNbottomAttachment, XmATTACH_NONE,
         XmNrightAttachment,  XmATTACH_WIDGET,
         XmNrightWidget,      objectSelect,
         XmNleftAttachment,   XmATTACH_NONE,
         XmNborderWidth,      2,
         XmNleftOffset,       5,
         XmNrightOffset,      5,
         XmNtopOffset,        5,
         XmNbottomOffset,     5,
         NULL);
         
   XtVaSetValues(objectSelect,
         XmNtopAttachment,    XmATTACH_FORM,
         XmNbottomAttachment, XmATTACH_NONE,
         XmNbottomWidget,     controls,
         XmNrightAttachment,  XmATTACH_FORM,
         XmNleftAttachment,   XmATTACH_NONE,
         XmNborderWidth,      2,
         XmNleftOffset,       5,
         XmNrightOffset,      5,
         XmNtopOffset,        5,
         XmNbottomOffset,     5,
         NULL);
   */      
   /*}}}  */
   
   XtRealizeWidget(xdata.topWidget);
   
   xdata.drawWin = XtWindow(drawArea);
   
   xdata.patternList = XmSelectionBoxGetChild(patternSelect, XmDIALOG_LIST);
   add_ss(cur_ss, labelStr);
   
   while( p = XtAppPending(xdata.appContext) )
      XtAppProcessEvent(xdata.appContext, p);
}  
/*}}}  */
   
JballEvent()
/*{{{  */
{
   XtInputMask pending;

   while( pending = XtAppPending(xdata.appContext) )
   {
      XtAppProcessEvent(xdata.appContext, pending);
   }  
   
   if (view_delta != D_NONE)
      flyButtonArmCallBack(NULL, (XtPointer)view_delta, NULL);

   if (hand_delta != D_NONE)
      handPosButtonArmCallBack(NULL, (XtPointer)hand_delta, NULL);
} 
/*}}}  */

JballEnd()
/*{{{  */
{
  XFreeGC( xdata.display, xdata.XorGC);
  XFreeGC( xdata.display, xdata.RevGC);
  XtDestroyApplicationContext(xdata.appContext);
} 
/*}}}  */

void
flush_graphics()
{
   XFlush(xdata.display);
}

getmaxx()
{
    return (STD_WIDTH);
}

getmaxy()
{
    return (STD_HEIGHT);
}

int kbhit()
/*{{{  */
{
    JballEvent();

    return( FALSE );
}
/*}}}  */

int getch()
/*{{{  */
{
    return(' ');
}
/*}}}  */

void closegraph(void)
/*{{{  */
{  
  JballEnd();
}
/*}}}  */

void clrscr(void)
/*{{{  */
{
   if (xdata.drawWin != NULL)
   {
      XClearWindow( xdata.display, xdata.drawWin );

      /* 
       * Oh God, this is horrible: update the state of any widgets whose
       * values may have been changed by the juggling loop, which changes 
       * the pattern asynchronously. It always does a redo_screen() when
       * this happens, though, so clrscr() should be called at the right time.
       */
       
      XtVaSetValues(xdata.detail,
                    XmNvalue, cur_ss->detail,
                    NULL);
      XtVaSetValues(xdata.height,
                    XmNvalue, cur_ss->height,
                    NULL);
      /*
      XtVaSetValues(xdata.scoop,
                    XmNvalue, cur_ss->scoop,
                    NULL);
      */
   }
}     
/*}}}  */

/* stub routines for unused IBMPC-style calls */
/* for speed, define these to be empty macros instead */

/*{{{  */
void cleardevice(void)
{}

void gotoxy(void)
{}

void grapherrormsg(void)
{}

void m_colormode(void)
{}

void setbkcolor(void)
{}
/*}}}  */

/*eof*/
